
#include "sdbInt.h"

#define TRUE  1
#define FALSE 0
#define EQUIVALENT 0

#define FIELD_SEP " \t\n\r"
#define RECORD_SEP "\n\r"

static int               SdbTextPosition _ANSI_ARGS_((char *list[], 
				CONST char *token, int k));
static int               SdbTextVar _ANSI_ARGS_((SdbTextFile *file, 
				char *string));
static int               SdbTextData _ANSI_ARGS_((SdbTextFile *file, 
				char *string));
static int               SdbTextNext _ANSI_ARGS_((SdbTextFile *file));

static int
SdbTextPosition(list, token, k)
     char                   *list[];
     CONST char             *token;
     int                     k;
{
    int                     i;

    for (i = 0; i < k; i++) {
        if (list[i] && (strcmp(list[i], token) == EQUIVALENT)) {
            break;
        }
    }
    return (i);
}

static int
SdbTextVar(file, string)
     SdbTextFile            *file;
     char                   *string;
{
    Tcl_Interp             *interp = file->interp;
    int                     i,
                            position,
                            k,
                            m,
                            split=-1,
                            t;
    char                  *input[SDB_MAX_FIELDS];
    int                     listC;
    char                  **listV;

    if (!file) {
        fprintf(stderr, "null text_file pointer\n");
        exit(1);
    }

    /*
     * parse the fields off of the VAR statement 
     */
    if (Tcl_SplitList(interp, string, &listC, &listV) != TCL_OK) {
        return (0);
    }

    for (m = 0, i = 0; i < listC; i++) {
        if (m >= SDB_MAX_FIELDS) {
            fprintf(stderr, "Maximum input fields exceeded\n");
            exit(1);
        }
        if (strcmp(listV[i], ":"))
            input[m++] = listV[i];
        else
            split = m;
    }
    file->m = m;
    /*
     * The variable split keeps track of where the ':' character
     * is found.  If there was no ':', then all of the fields are
     * code type.
     */
    if (split == -1) 
      split=m;

    /* 
     * If this is the first time this has been called, load up
     * the file structure with default values 
     */
    if (file->n == 0) {
        file->split = split;
        for (t = 0; t < m; t++) {
            file->remap[t] = t;
            if (t < file->split)
	      file->present[t] = SDB_CODE;
            else
	      file->present[t] = SDB_NUMBER;
	    file->field[t] = strdup(input[t]);
	}
        file->k = file->n = t;
    }

    /* if this is not the first time through (perhaps a second
     * VAR statement encountered?) then reprocess the VAR
     * string */
    else {
        k = 0;
	/* 
	 * go through the fields from the original segment one by one
	 */
        for (t = 0; t < file->n; t++) {

	    /* 
	     * check to see if this column existed in the original
	     * file segment.  If it does, set the remap column.
	     * Make sure that it is still the same type (code 
	     * or numeric).  If the type has changed then set it to
	     * missing.
	     */
            position = SdbTextPosition(input, file->field[t], m);
            if (position < m) {
                file->remap[t] = position;
		if ((file->present[t]&SDB_NUMBER) &&
		    (position < split))
		  file->present[t] |= SDB_MISSING;
		else if ((file->present[t]&SDB_CODE) &&
		    (position >= split))
		  file->present[t] |= SDB_MISSING;
		else
		  k++;
            } else {
	      file->present[t] |= SDB_MISSING;
	    }
        }
        file->k = k;
    }

    ckfree(listV);
    file->current = var;
    return (file->k);
}


static int
SdbTextData(file, string)
     SdbTextFile            *file;
     char                   *string;
{
    Tcl_Interp             *interp = file->interp;
    int                     position,
                            t,
                            m=0;
    char                   *input[SDB_MAX_FIELDS];
    static char            *missing = ".";
    int                     i,
                            listC;
    char                  **listV = file->listV;

    if (!file) {
        fprintf(stderr, "null text_file pointer\n");
        exit(1);
    }

    /*
     * free the old list contents, then split the input list.
     */
    if (listV)
        ckfree(listV);
    if (Tcl_SplitList(interp, string, &listC, &listV) != TCL_OK)
        return 0;
    file->listV = listV;

    /*
     * Copy to remove the record seperator
     */
    for (i = 0; i < listC; i++) {
        if (strcmp(":", listV[i]))
            input[m++] = listV[i];
    }
    file->itemsRead = m;
    for (; m < file->m; m++)
        input[m] = missing;

    /*
     * Now copy the pointers into the list structure.  This
     * takes into account missing values, reorderd columns, etc.
     */
    for (t = 0; t < file->n; t++) {
        if (file->present[t] & SDB_MISSING) {
            file->list[t] = missing;
        } else {
            position = file->remap[t];
            file->list[t] = input[position];
        }
    }

    file->current = data;
    return (m);
}

static int
SdbTextNext(file)
     SdbTextFile            *file;
{
    int                     k;
    char                   *string,
                           *field;

    /*
     * process the next record in the input stream 
     */
    k = 0;
    if (fgets(file->record, SDB_RECORD_LENGTH, file->stream)) {
        if ((field = strtok(file->record, FIELD_SEP))) {
            string = strtok(NULL, RECORD_SEP);
            if (strcmp(field, "VAR") == 0)
                k = SdbTextVar(file, string);
            else if (strcmp(field, "DATA") == 0)
                k = SdbTextData(file, string);
            else
                file->current = undefined;
        }
    } else {
        file->current = end;
        file->active = FALSE;
    }
    return (k);
}

SdbTextFile            *
SdbTextOpen(interp, string, size)
     Tcl_Interp             *interp;
     char                   *string;
     int                     size;
{
    FILE                   *stream;
    SdbTextFile            *file;
    int                     i;

    /*
     * Parse the arguments
     */
    if (string) {
      if ((stream = fopen(string, "r"))) {
      } else {
	interp->result = "Could not open file";
	return (NULL);
      }
    } else {
      interp->result = "Bad file name";
      return (NULL);
    }

    /* 
     * set size of input buffer 
     */
    if (size > 0)
        setvbuf(stream, NULL, _IOFBF, (unsigned) size);

    /*
     * allocate space for the input file and initialize
     * the fields.
     */
    if (!(file = (SdbTextFile *) ckalloc(sizeof(SdbTextFile)))) {
        interp->result = "Out of memory";
        return NULL;
    }    
    file->active = TRUE;
    file->stream = stream;
    file->current = undefined;
    file->interp = interp;
    file->listV = NULL;
    file->n = 0;
    file->k = 0;
    file->m = 0;
    file->split = 0;
    for (i=0;i<SDB_MAX_FIELDS;i++) {
      file->list[i]=NULL;
      file->field[i]=NULL;
      file->present[i]=SDB_MISSING;
      file->remap[i]=0;
    }

    /*
     * Scan the file for the first VAR statement
     */
    do {
        i = SdbTextNext(file);
    } while (file->current != var && file->current != end);

    if (file->current == end) {
        SdbTextClose(file);
        interp->result = "No VAR statement in file";
        return (NULL);
    }
    return (file);
}

void
SdbTextClose(file)
     SdbTextFile            *file;
{
    int                     i;

    /*
     * Close the file and free the storage.
     */
    if (!file) {
        fprintf(stderr, "null text_file pointer");
        exit(1);
    }
    if (file->stream)
        fclose(file->stream);

    for (i = 0; i < file->n; i++) {
        if (file->field[i])
            ckfree(file->field[i]);
    }
    if (file->listV)
        ckfree(file->listV);

    ckfree(file);
}

int
SdbTextGet(file)
     SdbTextFile            *file;
{
    int                     k;

    if (!file) {
        fprintf(stderr, "null text_file pointer\n");
        exit(1);
    }
    if (!file->active)
        return (0);
    do {
        k = SdbTextNext(file);
    } while ((file->current != data) && (file->current != end));

    if (file->current == data)
        return (TRUE);
    else
        return (FALSE);
}

int
SdbTextReposition(file, field, new_pos)
     SdbTextFile            *file;
     char                   *field;
     int                     new_pos;
{
    /* this routine will shuffle the order of returned variables.
     * If a variable to be shuffled does not exist, if will
     * be inserted as a missing variable. */

    char                   *cswap;
    int                     iswap,
                            old_pos,
                            i;

    if (!file) {
        fprintf(stderr, "null text_file pointer\n");
        exit(1);
    }
    /* First handle the case where the variable exists */
    if ((old_pos = SdbTextPosition(file->field, field, file->n)) < file->n) {
        iswap = file->remap[new_pos];
        file->remap[new_pos] = file->remap[old_pos];
        file->remap[old_pos] = iswap;
        cswap = file->field[new_pos];
        file->field[new_pos] = file->field[old_pos];
        file->field[old_pos] = cswap;
        iswap = file->present[new_pos];
        file->present[new_pos] = file->present[old_pos];
        file->present[old_pos] = iswap;
        return (TRUE);
    }
    /* Now handle the case where the variable does not exist */
    else {
        if (file->n >= SDB_MAX_FIELDS) {
            fprintf(stderr, "Too many variables in file table\n");
            exit(1);
        }
        for (i = file->n; i >= new_pos; i--) {
            file->field[i + 1] = file->field[i];
            file->remap[i + 1] = file->remap[i];
            file->present[i + 1] = file->present[i];
        }
        if (!(file->field[new_pos] = (char *) ckalloc(strlen(field) + 1))) {
            fprintf(stderr, "Out of memory");
            exit(1);
        }
        strcpy(file->field[new_pos], field);
        file->present[new_pos] = SDB_MISSING;
        file->remap[new_pos] = 0;
        file->n++;
        return (FALSE);
    }
}
